#!/usr/bin/env ruby
require 'faye'
require 'eventmachine'
require 'rest-client'

require 'json'
require 'logger'

# Connects to a partygoer instance, performs an initial poll, and then polls
# every 5 minutes. In between polls the client stays in sync using faye to
# recieve push updates
class PartyGoerClient
  attr_reader :queued, :playing, :played

  @@logger = Logger.new('partygoerclient.log', 2, 1_000_000)
  @@logger.formatter = proc do |severity, datetime, _progname, msg|
    "\n\n#{datetime} #{severity}:\n\t#{msg}"
  end
  @@logger.level = Logger::INFO

  def initialize(server)
    @server = server

    @played = []
    @queued = []
    @playing = nil

    @played_lock = Mutex.new
    @queued_lock = Mutex.new
    @playing_lock = Mutex.new

    @event_loop = nil

    @api = RestClient::Resource.new("#{@server}/api/queue", headers:
               { content_type: 'application/json', accept: 'application/json' })
    @client = Faye::Client.new("#{@server}:9292/faye")
    logger.info { "Client initialized with server: #{server}" }
    @is_playing = false
    start
  end

  def playing=(track)
    payload = { track: track }
    @api['/playing'].put(payload)
    logger.info { "Set now playing to #{track['name']}" }
  rescue
    logger.error {
      "Could not set now playing to #{track.inspect}:\n\t#{$!}\n\t#{$@.join("\n\t")}"
    }
  end

  def suggest(track)
    case track
    when String
      payload = { track: { spotify_uri: track } }
    when Hash
      payload = { track: track }
    end
    EM.defer(proc { @api['/tracks'].post(payload) })
    logger.info { "Suggested #{track.inspect}" }
  rescue
    logger.error("Failed to suggest track: #{track.inspect}\n#{$!}\n#{$@.join("\n\t")}")
  end

  def skip?
    if @playing['num_upvotes'] == 0
      @playing['num_downvotes'] >= 2
    else
      @playing['num_downvotes'] >= (2 * @playing['num_upvotes'])
    end
  rescue
    logger.error("Failed to check skip criteria:\n#{$!}\n#{$@.join("\n\t")}")
  end

  def up_next
    @queued[1]
  rescue
    logger.error("Failed to check up next:\n#{$!}\n#{$@.join("\n\t")}")
  end

  def start
    pull_update
    return if @event_loop
    logger.info { 'Starting event loop' }
    start_event_loop

    at_exit do
      stop
    end
  end

  def stop
    return unless @event_loop
    logger.info { 'Stopping event loop' }
    EM.stop_event_loop
    @event_loop.kill
    @event_loop = nil
  end

  # Aliases
  def now_playing
    playing
  end

  def currently_queued
    queued
  end

  def recently_played
    played
  end

  def logger
    @@logger
  end

  private
    def start_event_loop
      @event_loop = Thread.new do
        begin
          EM.run do
            EM.add_periodic_timer(300) { EM.defer { pull_update } }

            @client.subscribe('/track/vote') do |message|
              EM.defer(on_voted(message))
            end

            @client.subscribe('/track/create') do |message|
              EM.defer(on_created(message))
            end

            @client.subscribe('/queue/next') do |message|
              EM.defer(on_next(message))
            end
          end
        rescue
          logger.critical { "Could not start event loop:\n\t#{$!}\n\t#{$@.join("\n\t")}" }
          stop
          raise $!
        end
      end
    rescue
      logger.critical { "Could not start event loop:\n\t#{$!}\n\t#{$@.join("\n\t")}" }
      stop
      raise $!
    end


    def on_voted(track)
      proc {
        logger.debug { "Track voted on: #{track.inspect}" }
        begin
          @queued_lock.synchronize{
            index = @queued.find_index {
              |item| item['id'] == track['id']
            }
            @queued[index] = track
          }
        rescue
          logger.error { "Error in on_voted:\n\t#{$!}\n\t#{$@.join("\n\t")}" }
        ensure
          sort!
        end
      }
    end

    def on_created(track)
      proc {
        logger.debug { "Track created: #{track.inspect}" }
        begin
          @queued_lock.synchronize{
            @queued << track
          }
        rescue
          logger.error { "Error in on_created:\n\t#{$!}\n\t#{$@.join("\n\t")}" }
        ensure
          sort!
        end
      }
    end

    def on_next(track)
      proc {
        logger.debug { "Next track called: #{track.inspect}" }
        begin
          @playing_lock.synchronize {
            @playing['playing'] = false
            @played_lock.synchronize { @played << @playing }
            @queued_lock.synchronize {
              @queued.delete_if {
                 |item| item['id'] == @playing['id']
              }
              index = @queued.find_index {
                |item| item['id'] == track['id']
              }
              @queued[index] = track
            }
            @playing = track
          }
        rescue
          logger.error { "Error in on_next:\n\t#{$!}\n\t#{$@.join("\n\t")}" }
        ensure
          sort!
        end
      }
    end

    def pull_update
      logger.info { "Polling for update" }
      begin
        body = JSON.parse(@api['/recently_played'].get.body)
        @played_lock.synchronize { @played = body }
      rescue
        logger.error { "Could not fetch recently played:\n\t#{$!}\n\t#{$@.join("\n\t")}" }
      end
      begin
        body = JSON.parse(@api['/tracks'].get.body)
        @queued_lock.synchronize { @queued = body }
      rescue
        logger.error { "Could not fetch queue:\n\t#{$!}\n\t#{$@.join("\n\t")}" }
      end
      begin
        body = JSON.parse(@api['/playing'].get.body)
        @playing_lock.synchronize { @playing = body }
      rescue
        logger.error { "Could not fetch now playing:\n\t#{$!}\n\t#{$@.join("\n\t")}" }
      end
      sort!
    end

    def sort!
      begin
        @played_lock.synchronize{
          @played.sort_by! { |track| track['id'] }
        }
        logger.debug { 'Sorted recently played' }
      rescue
        logger.error { "Could not sort recently played:\n\t#{$!}\n\t#{$@.join("\n\t")}" }
      end
      begin
        @queued_lock.synchronize{
          @queued.sort_by! { |track|
            i = track['playing'] ? 0 : 1
            [i, -1 * track['score'], track['id']]
          }
        }
        logger.debug { 'Sorted queued' }
      rescue
        logger.error { "Could not sort the queue:\n\t#{$!}\n\t#{$@.join("\n\t")}" }
      end
      begin
        @playing_lock.synchronize{
          @playing = @queued.select { |track| track['playing'] }.first
        }
        logger.debug { 'Updated now playing' }
      rescue
        logger.error { "Could not update now playing:\n\t#{$!}\n\t#{$@.join("\n\t")}" }
      end
    end
end
